# Copyright (C) 2012-2020 Ben Kurtovic <ben.kurtovic@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Test cases for the Token class and its subclasses.
"""

import pytest

from mwparserfromhell.parser import tokens


@pytest.mark.parametrize("name", tokens.__all__)
def test_issubclass(name):
    """check that all classes within the tokens module are really Tokens"""
    klass = getattr(tokens, name)
    assert issubclass(klass, tokens.Token) is True
    assert isinstance(klass(), klass)
    assert isinstance(klass(), tokens.Token)


def test_attributes():
    """check that Token attributes can be managed properly"""
    token1 = tokens.Token()
    token2 = tokens.Token(foo="bar", baz=123)

    assert "bar" == token2.foo
    assert 123 == token2.baz
    assert token1.foo is None
    assert token2.bar is None

    token1.spam = "eggs"
    token2.foo = "ham"
    del token2.baz

    assert "eggs" == token1.spam
    assert "ham" == token2.foo
    assert token2.baz is None
    with pytest.raises(KeyError):
        token2.__delattr__("baz")


def test_repr():
    """check that repr() on a Token works as expected"""
    token1 = tokens.Token()
    token2 = tokens.Token(foo="bar", baz=123)
    token3 = tokens.Text(text="earwig" * 100)
    hundredchars = ("earwig" * 100)[:97] + "..."

    assert "Token()" == repr(token1)
    assert repr(token2) in ("Token(foo='bar', baz=123)", "Token(baz=123, foo='bar')")
    assert "Text(text='" + hundredchars + "')" == repr(token3)


def test_equality():
    """check that equivalent tokens are considered equal"""
    token1 = tokens.Token()
    token2 = tokens.Token()
    token3 = tokens.Token(foo="bar", baz=123)
    token4 = tokens.Text(text="asdf")
    token5 = tokens.Text(text="asdf")
    token6 = tokens.TemplateOpen(text="asdf")

    assert token1 == token2
    assert token2 == token1
    assert token4 == token5
    assert token5 == token4
    assert token1 != token3
    assert token2 != token3
    assert token4 != token6
    assert token5 != token6


@pytest.mark.parametrize(
    "token",
    [tokens.Token(), tokens.Token(foo="bar", baz=123), tokens.Text(text="earwig")],
)
def test_repr_equality(token):
    """check that eval(repr(token)) == token"""
    assert token == eval(repr(token), vars(tokens))



# Copyright (C) 2012-2016 Ben Kurtovic <ben.kurtovic@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Tests for the Parser class itself, which tokenizes and builds nodes.
"""

import pytest

from mwparserfromhell import parser
from mwparserfromhell.nodes import Tag, Template, Text, Wikilink
from mwparserfromhell.nodes.extras import Parameter
from .conftest import assert_wikicode_equal, wrap, wraptext


@pytest.fixture()
def pyparser():
    """make sure the correct tokenizer is used"""
    restore = parser.use_c
    if parser.use_c:
        parser.use_c = False
    yield
    parser.use_c = restore


def test_use_c(pyparser):
    assert parser.Parser()._tokenizer.USES_C is False


def test_parsing(pyparser):
    """integration test for parsing overall"""
    text = "this is text; {{this|is=a|template={{with|[[links]]|in}}it}}"
    expected = wrap(
        [
            Text("this is text; "),
            Template(
                wraptext("this"),
                [
                    Parameter(wraptext("is"), wraptext("a")),
                    Parameter(
                        wraptext("template"),
                        wrap(
                            [
                                Template(
                                    wraptext("with"),
                                    [
                                        Parameter(
                                            wraptext("1"),
                                            wrap([Wikilink(wraptext("links"))]),
                                            showkey=False,
                                        ),
                                        Parameter(
                                            wraptext("2"), wraptext("in"), showkey=False
                                        ),
                                    ],
                                ),
                                Text("it"),
                            ]
                        ),
                    ),
                ],
            ),
        ]
    )
    actual = parser.Parser().parse(text)
    assert_wikicode_equal(expected, actual)


def test_skip_style_tags(pyparser):
    """test Parser.parse(skip_style_tags=True)"""
    text = "This is an example with ''italics''!"
    a = wrap(
        [
            Text("This is an example with "),
            Tag(wraptext("i"), wraptext("italics"), wiki_markup="''"),
            Text("!"),
        ]
    )
    b = wraptext("This is an example with ''italics''!")

    with_style = parser.Parser().parse(text, skip_style_tags=False)
    without_style = parser.Parser().parse(text, skip_style_tags=True)
    assert_wikicode_equal(a, with_style)
    assert_wikicode_equal(b, without_style)



# Copyright (C) 2012-2020 Ben Kurtovic <ben.kurtovic@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Test cases for the Comment node.
"""

import pytest

from mwparserfromhell.nodes import Comment


def test_str():
    """test Comment.__str__()"""
    node = Comment("foobar")
    assert "<!--foobar-->" == str(node)


def test_children():
    """test Comment.__children__()"""
    node = Comment("foobar")
    gen = node.__children__()
    with pytest.raises(StopIteration):
        next(gen)


def test_strip():
    """test Comment.__strip__()"""
    node = Comment("foobar")
    assert node.__strip__() is None


def test_showtree():
    """test Comment.__showtree__()"""
    output = []
    node = Comment("foobar")
    node.__showtree__(output.append, None, None)
    assert ["<!--foobar-->"] == output


def test_contents():
    """test getter/setter for the contents attribute"""
    node = Comment("foobar")
    assert "foobar" == node.contents
    node.contents = "barfoo"
    assert "barfoo" == node.contents



# Copyright (C) 2012-2020 Ben Kurtovic <ben.kurtovic@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Test cases for the Wikilink node.
"""

import pytest

from mwparserfromhell.nodes import Text, Wikilink
from .conftest import assert_wikicode_equal, wrap, wraptext


def test_str():
    """test Wikilink.__str__()"""
    node = Wikilink(wraptext("foobar"))
    assert "[[foobar]]" == str(node)
    node2 = Wikilink(wraptext("foo"), wraptext("bar"))
    assert "[[foo|bar]]" == str(node2)


def test_children():
    """test Wikilink.__children__()"""
    node1 = Wikilink(wraptext("foobar"))
    node2 = Wikilink(wraptext("foo"), wrap([Text("bar"), Text("baz")]))
    gen1 = node1.__children__()
    gen2 = node2.__children__()
    assert node1.title == next(gen1)
    assert node2.title == next(gen2)
    assert node2.text == next(gen2)
    with pytest.raises(StopIteration):
        next(gen1)
    with pytest.raises(StopIteration):
        next(gen2)


def test_strip():
    """test Wikilink.__strip__()"""
    node = Wikilink(wraptext("foobar"))
    node2 = Wikilink(wraptext("foo"), wraptext("bar"))
    assert "foobar" == node.__strip__()
    assert "bar" == node2.__strip__()


def test_showtree():
    """test Wikilink.__showtree__()"""
    output = []
    getter, marker = object(), object()
    get = lambda code: output.append((getter, code))
    mark = lambda: output.append(marker)
    node1 = Wikilink(wraptext("foobar"))
    node2 = Wikilink(wraptext("foo"), wraptext("bar"))
    node1.__showtree__(output.append, get, mark)
    node2.__showtree__(output.append, get, mark)
    valid = [
        "[[",
        (getter, node1.title),
        "]]",
        "[[",
        (getter, node2.title),
        "    | ",
        marker,
        (getter, node2.text),
        "]]",
    ]
    assert valid == output


def test_title():
    """test getter/setter for the title attribute"""
    title = wraptext("foobar")
    node1 = Wikilink(title)
    node2 = Wikilink(title, wraptext("baz"))
    assert title is node1.title
    assert title is node2.title
    node1.title = "héhehé"
    node2.title = "héhehé"
    assert_wikicode_equal(wraptext("héhehé"), node1.title)
    assert_wikicode_equal(wraptext("héhehé"), node2.title)


def test_text():
    """test getter/setter for the text attribute"""
    text = wraptext("baz")
    node1 = Wikilink(wraptext("foobar"))
    node2 = Wikilink(wraptext("foobar"), text)
    assert None is node1.text
    assert text is node2.text
    node1.text = "buzz"
    node2.text = None
    assert_wikicode_equal(wraptext("buzz"), node1.text)
    assert None is node2.text



# Copyright (C) 2012-2020 Ben Kurtovic <ben.kurtovic@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Test cases for the Template node.
"""

from difflib import unified_diff

import pytest

from mwparserfromhell.nodes import HTMLEntity, Template, Text
from mwparserfromhell.nodes.extras import Parameter
from mwparserfromhell import parse
from .conftest import assert_wikicode_equal, wrap, wraptext

pgens = lambda k, v: Parameter(wraptext(k), wraptext(v), showkey=True)
pgenh = lambda k, v: Parameter(wraptext(k), wraptext(v), showkey=False)


def test_str():
    """test Template.__str__()"""
    node = Template(wraptext("foobar"))
    assert "{{foobar}}" == str(node)
    node2 = Template(wraptext("foo"), [pgenh("1", "bar"), pgens("abc", "def")])
    assert "{{foo|bar|abc=def}}" == str(node2)


def test_children():
    """test Template.__children__()"""
    node2p1 = Parameter(wraptext("1"), wraptext("bar"), showkey=False)
    node2p2 = Parameter(wraptext("abc"), wrap([Text("def"), Text("ghi")]), showkey=True)
    node1 = Template(wraptext("foobar"))
    node2 = Template(wraptext("foo"), [node2p1, node2p2])

    gen1 = node1.__children__()
    gen2 = node2.__children__()
    assert node1.name == next(gen1)
    assert node2.name == next(gen2)
    assert node2.params[0].value == next(gen2)
    assert node2.params[1].name == next(gen2)
    assert node2.params[1].value == next(gen2)
    with pytest.raises(StopIteration):
        next(gen1)
    with pytest.raises(StopIteration):
        next(gen2)


def test_strip():
    """test Template.__strip__()"""
    node1 = Template(wraptext("foobar"))
    node2 = Template(
        wraptext("foo"), [pgenh("1", "bar"), pgens("foo", ""), pgens("abc", "def")]
    )
    node3 = Template(
        wraptext("foo"),
        [
            pgenh("1", "foo"),
            Parameter(
                wraptext("2"), wrap([Template(wraptext("hello"))]), showkey=False
            ),
            pgenh("3", "bar"),
        ],
    )

    assert node1.__strip__(keep_template_params=False) is None
    assert node2.__strip__(keep_template_params=False) is None
    assert "" == node1.__strip__(keep_template_params=True)
    assert "bar def" == node2.__strip__(keep_template_params=True)
    assert "foo bar" == node3.__strip__(keep_template_params=True)


def test_showtree():
    """test Template.__showtree__()"""
    output = []
    getter, marker = object(), object()
    get = lambda code: output.append((getter, code))
    mark = lambda: output.append(marker)
    node1 = Template(wraptext("foobar"))
    node2 = Template(wraptext("foo"), [pgenh("1", "bar"), pgens("abc", "def")])
    node1.__showtree__(output.append, get, mark)
    node2.__showtree__(output.append, get, mark)
    valid = [
        "{{",
        (getter, node1.name),
        "}}",
        "{{",
        (getter, node2.name),
        "    | ",
        marker,
        (getter, node2.params[0].name),
        "    = ",
        marker,
        (getter, node2.params[0].value),
        "    | ",
        marker,
        (getter, node2.params[1].name),
        "    = ",
        marker,
        (getter, node2.params[1].value),
        "}}",
    ]
    assert valid == output


def test_name():
    """test getter/setter for the name attribute"""
    name = wraptext("foobar")
    node1 = Template(name)
    node2 = Template(name, [pgenh("1", "bar")])
    assert name is node1.name
    assert name is node2.name
    node1.name = "asdf"
    node2.name = "téstïng"
    assert_wikicode_equal(wraptext("asdf"), node1.name)
    assert_wikicode_equal(wraptext("téstïng"), node2.name)


def test_params():
    """test getter for the params attribute"""
    node1 = Template(wraptext("foobar"))
    plist = [pgenh("1", "bar"), pgens("abc", "def")]
    node2 = Template(wraptext("foo"), plist)
    assert [] == node1.params
    assert plist is node2.params


def test_has():
    """test Template.has()"""
    node1 = Template(wraptext("foobar"))
    node2 = Template(wraptext("foo"), [pgenh("1", "bar"), pgens("\nabc ", "def")])
    node3 = Template(
        wraptext("foo"), [pgenh("1", "a"), pgens("b", "c"), pgens("1", "d")]
    )
    node4 = Template(wraptext("foo"), [pgenh("1", "a"), pgens("b", " ")])
    assert node1.has("foobar", False) is False
    assert node2.has(1, False) is True
    assert node2.has("abc", False) is True
    assert node2.has("def", False) is False
    assert node3.has("1", False) is True
    assert node3.has(" b ", False) is True
    assert node4.has("b", False) is True
    assert node3.has("b", True) is True
    assert node4.has("b", True) is False
    assert node1.has_param("foobar", False) is False
    assert node2.has_param(1, False) is True


def test_get():
    """test Template.get()"""
    node1 = Template(wraptext("foobar"))
    node2p1 = pgenh("1", "bar")
    node2p2 = pgens("abc", "def")
    node2 = Template(wraptext("foo"), [node2p1, node2p2])
    node3p1 = pgens("b", "c")
    node3p2 = pgens("1", "d")
    node3 = Template(wraptext("foo"), [pgenh("1", "a"), node3p1, node3p2])
    node4p1 = pgens(" b", " ")
    node4 = Template(wraptext("foo"), [pgenh("1", "a"), node4p1])
    with pytest.raises(ValueError):
        node1.get("foobar")
    assert node2p1 is node2.get(1)
    assert node2p2 is node2.get("abc")
    with pytest.raises(ValueError):
        node2.get("def")
    assert node3p1 is node3.get("b")
    assert node3p2 is node3.get("1")
    assert node4p1 is node4.get("b ")


def test_add():
    """test Template.add()"""
    node1 = Template(wraptext("a"), [pgens("b", "c"), pgenh("1", "d")])
    node2 = Template(wraptext("a"), [pgens("b", "c"), pgenh("1", "d")])
    node3 = Template(wraptext("a"), [pgens("b", "c"), pgenh("1", "d")])
    node4 = Template(wraptext("a"), [pgens("b", "c"), pgenh("1", "d")])
    node5 = Template(wraptext("a"), [pgens("b", "c"), pgens("    d ", "e")])
    node6 = Template(wraptext("a"), [pgens("b", "c"), pgens("b", "d"), pgens("b", "e")])
    node7 = Template(wraptext("a"), [pgens("b", "c"), pgenh("1", "d")])
    node8p = pgenh("1", "d")
    node8 = Template(wraptext("a"), [pgens("b", "c"), node8p])
    node9 = Template(wraptext("a"), [pgens("b", "c"), pgenh("1", "d")])
    node10 = Template(wraptext("a"), [pgens("b", "c"), pgenh("1", "e")])
    node11 = Template(wraptext("a"), [pgens("b", "c")])
    node12 = Template(wraptext("a"), [pgens("b", "c")])
    node13 = Template(
        wraptext("a"), [pgens("\nb ", " c"), pgens("\nd ", " e"), pgens("\nf ", " g")]
    )
    node14 = Template(
        wraptext("a\n"),
        [
            pgens("b ", "c\n"),
            pgens("d ", " e"),
            pgens("f ", "g\n"),
            pgens("h ", " i\n"),
        ],
    )
    node15 = Template(
        wraptext("a"),
        [pgens("b  ", " c\n"), pgens("\nd  ", " e"), pgens("\nf  ", "g ")],
    )
    node16 = Template(
        wraptext("a"), [pgens("\nb ", " c"), pgens("\nd ", " e"), pgens("\nf ", " g")]
    )
    node17 = Template(wraptext("a"), [pgenh("1", "b")])
    node18 = Template(wraptext("a"), [pgenh("1", "b")])
    node19 = Template(wraptext("a"), [pgenh("1", "b")])
    node20 = Template(
        wraptext("a"),
        [pgenh("1", "b"), pgenh("2", "c"), pgenh("3", "d"), pgenh("4", "e")],
    )
    node21 = Template(
        wraptext("a"),
        [pgenh("1", "b"), pgenh("2", "c"), pgens("4", "d"), pgens("5", "e")],
    )
    node22 = Template(
        wraptext("a"),
        [pgenh("1", "b"), pgenh("2", "c"), pgens("4", "d"), pgens("5", "e")],
    )
    node23 = Template(wraptext("a"), [pgenh("1", "b")])
    node24 = Template(wraptext("a"), [pgenh("1", "b")])
    node25 = Template(wraptext("a"), [pgens("b", "c")])
    node26 = Template(wraptext("a"), [pgenh("1", "b")])
    node27 = Template(wraptext("a"), [pgenh("1", "b")])
    node28 = Template(wraptext("a"), [pgens("1", "b")])
    node29 = Template(
        wraptext("a"), [pgens("\nb ", " c"), pgens("\nd ", " e"), pgens("\nf ", " g")]
    )
    node30 = Template(
        wraptext("a\n"),
        [
            pgens("b ", "c\n"),
            pgens("d ", " e"),
            pgens("f ", "g\n"),
            pgens("h ", " i\n"),
        ],
    )
    node31 = Template(
        wraptext("a"),
        [pgens("b  ", " c\n"), pgens("\nd  ", " e"), pgens("\nf  ", "g ")],
    )
    node32 = Template(
        wraptext("a"),
        [pgens("\nb ", " c "), pgens("\nd ", " e "), pgens("\nf ", " g ")],
    )
    node33 = Template(
        wraptext("a"),
        [
            pgens("b", "c"),
            pgens("d", "e"),
            pgens("b", "f"),
            pgens("b", "h"),
            pgens("i", "j"),
        ],
    )
    node34 = Template(
        wraptext("a"),
        [pgens("1", "b"), pgens("x", "y"), pgens("1", "c"), pgens("2", "d")],
    )
    node35 = Template(
        wraptext("a"),
        [pgens("1", "b"), pgens("x", "y"), pgenh("1", "c"), pgenh("2", "d")],
    )
    node36 = Template(
        wraptext("a"), [pgens("b", "c"), pgens("d", "e"), pgens("f", "g")]
    )
    node37 = Template(wraptext("a"), [pgenh("1", "")])
    node38 = Template(wraptext("abc"))
    node39 = Template(wraptext("a"), [pgenh("1", " b ")])
    node40 = Template(wraptext("a"), [pgenh("1", " b"), pgenh("2", " c")])
    node41 = Template(wraptext("a"), [pgens("1", " b"), pgens("2", " c")])
    node42 = Template(wraptext("a"), [pgens("b", "  \n")])

    node1.add("e", "f", showkey=True)
    node2.add(2, "g", showkey=False)
    node3.add("e", "foo|bar", showkey=True)
    node4.add("e", "f", showkey=True, before="b")
    node5.add("f", "g", showkey=True, before=" d     ")
    node6.add("f", "g", showkey=True, before="b")
    with pytest.raises(ValueError):
        node7.add("e", "f", showkey=True, before="q")
    node8.add("e", "f", showkey=True, before=node8p)
    node9.add("e", "f", showkey=True, before=pgenh("1", "d"))
    with pytest.raises(ValueError):
        node10.add("e", "f", showkey=True, before=pgenh("1", "d"))
    node11.add("d", "foo=bar", showkey=True)
    node12.add("1", "foo=bar", showkey=False)
    node13.add("h", "i", showkey=True)
    node14.add("j", "k", showkey=True)
    node15.add("h", "i", showkey=True)
    node16.add("h", "i", showkey=True, preserve_spacing=False)
    node17.add("2", "c")
    node18.add("3", "c")
    node19.add("c", "d")
    node20.add("5", "f")
    node21.add("3", "f")
    node22.add("6", "f")
    node23.add("c", "foo=bar")
    node24.add("2", "foo=bar")
    node25.add("b", "d")
    node26.add("1", "foo=bar")
    node27.add("1", "foo=bar", showkey=True)
    node28.add("1", "foo=bar", showkey=False)
    node29.add("d", "foo")
    node30.add("f", "foo")
    node31.add("f", "foo")
    node32.add("d", "foo", preserve_spacing=False)
    node33.add("b", "k")
    node34.add("1", "e")
    node35.add("1", "e")
    node36.add("d", "h", before="b")
    node37.add(1, "b")
    node38.add("1", "foo")
    with pytest.raises(ValueError):
        node38.add("z", "bar", showkey=False)
    node39.add("1", "c")
    node40.add("3", "d")
    node41.add("3", "d")
    node42.add("b", "hello")

    assert "{{a|b=c|d|e=f}}" == node1
    assert "{{a|b=c|d|g}}" == node2
    assert "{{a|b=c|d|e=foo&#124;bar}}" == node3
    assert isinstance(node3.params[2].value.get(1), HTMLEntity)
    assert "{{a|e=f|b=c|d}}" == node4
    assert "{{a|b=c|f=g|    d =e}}" == node5
    assert "{{a|b=c|b=d|f=g|b=e}}" == node6
    assert "{{a|b=c|d}}" == node7
    assert "{{a|b=c|e=f|d}}" == node8
    assert "{{a|b=c|e=f|d}}" == node9
    assert "{{a|b=c|e}}" == node10
    assert "{{a|b=c|d=foo=bar}}" == node11
    assert "{{a|b=c|foo&#61;bar}}" == node12
    assert isinstance(node12.params[1].value.get(1), HTMLEntity)
    assert "{{a|\nb = c|\nd = e|\nf = g|\nh = i}}" == node13
    assert "{{a\n|b =c\n|d = e|f =g\n|h = i\n|j =k\n}}" == node14
    assert "{{a|b  = c\n|\nd  = e|\nf  =g |\nh  = i}}" == node15
    assert "{{a|\nb = c|\nd = e|\nf = g|h=i}}" == node16
    assert "{{a|b|c}}" == node17
    assert "{{a|b|3=c}}" == node18
    assert "{{a|b|c=d}}" == node19
    assert "{{a|b|c|d|e|f}}" == node20
    assert "{{a|b|c|4=d|5=e|f}}" == node21
    assert "{{a|b|c|4=d|5=e|6=f}}" == node22
    assert "{{a|b|c=foo=bar}}" == node23
    assert "{{a|b|foo&#61;bar}}" == node24
    assert isinstance(node24.params[1].value.get(1), HTMLEntity)
    assert "{{a|b=d}}" == node25
    assert "{{a|foo&#61;bar}}" == node26
    assert isinstance(node26.params[0].value.get(1), HTMLEntity)
    assert "{{a|1=foo=bar}}" == node27
    assert "{{a|foo&#61;bar}}" == node28
    assert isinstance(node28.params[0].value.get(1), HTMLEntity)
    assert "{{a|\nb = c|\nd = foo|\nf = g}}" == node29
    assert "{{a\n|b =c\n|d = e|f =foo\n|h = i\n}}" == node30
    assert "{{a|b  = c\n|\nd  = e|\nf  =foo }}" == node31
    assert "{{a|\nb = c |\nd =foo|\nf = g }}" == node32
    assert "{{a|b=k|d=e|i=j}}" == node33
    assert "{{a|1=e|x=y|2=d}}" == node34
    assert "{{a|x=y|e|d}}" == node35
    assert "{{a|b=c|d=h|f=g}}" == node36
    assert "{{a|b}}" == node37
    assert "{{abc|foo}}" == node38
    assert "{{a|c}}" == node39
    assert "{{a| b| c|d}}" == node40
    assert "{{a|1= b|2= c|3= d}}" == node41
    assert "{{a|b=hello  \n}}" == node42


def test_remove():
    """test Template.remove()"""
    node1 = Template(wraptext("foobar"))
    node2 = Template(wraptext("foo"), [pgenh("1", "bar"), pgens("abc", "def")])
    node3 = Template(wraptext("foo"), [pgenh("1", "bar"), pgens("abc", "def")])
    node4 = Template(wraptext("foo"), [pgenh("1", "bar"), pgenh("2", "baz")])
    node5 = Template(
        wraptext("foo"), [pgens(" a", "b"), pgens("b", "c"), pgens("a ", "d")]
    )
    node6 = Template(
        wraptext("foo"), [pgens(" a", "b"), pgens("b", "c"), pgens("a ", "d")]
    )
    node7 = Template(
        wraptext("foo"), [pgens("1  ", "a"), pgens("  1", "b"), pgens("2", "c")]
    )
    node8 = Template(
        wraptext("foo"), [pgens("1  ", "a"), pgens("  1", "b"), pgens("2", "c")]
    )
    node9 = Template(
        wraptext("foo"), [pgens("1  ", "a"), pgenh("1", "b"), pgenh("2", "c")]
    )
    node10 = Template(
        wraptext("foo"), [pgens("1  ", "a"), pgenh("1", "b"), pgenh("2", "c")]
    )
    node11 = Template(
        wraptext("foo"), [pgens(" a", "b"), pgens("b", "c"), pgens("a ", "d")]
    )
    node12 = Template(
        wraptext("foo"), [pgens(" a", "b"), pgens("b", "c"), pgens("a ", "d")]
    )
    node13 = Template(
        wraptext("foo"), [pgens(" a", "b"), pgens("b", "c"), pgens("a ", "d")]
    )
    node14 = Template(
        wraptext("foo"), [pgens(" a", "b"), pgens("b", "c"), pgens("a ", "d")]
    )
    node15 = Template(
        wraptext("foo"), [pgens(" a", "b"), pgens("b", "c"), pgens("a ", "d")]
    )
    node16 = Template(
        wraptext("foo"), [pgens(" a", "b"), pgens("b", "c"), pgens("a ", "d")]
    )
    node17 = Template(
        wraptext("foo"), [pgens("1  ", "a"), pgenh("1", "b"), pgenh("2", "c")]
    )
    node18 = Template(
        wraptext("foo"), [pgens("1  ", "a"), pgenh("1", "b"), pgenh("2", "c")]
    )
    node19 = Template(
        wraptext("foo"), [pgens("1  ", "a"), pgenh("1", "b"), pgenh("2", "c")]
    )
    node20 = Template(
        wraptext("foo"), [pgens("1  ", "a"), pgenh("1", "b"), pgenh("2", "c")]
    )
    node21 = Template(
        wraptext("foo"),
        [
            pgens("a", "b"),
            pgens("c", "d"),
            pgens("e", "f"),
            pgens("a", "b"),
            pgens("a", "b"),
        ],
    )
    node22 = Template(
        wraptext("foo"),
        [
            pgens("a", "b"),
            pgens("c", "d"),
            pgens("e", "f"),
            pgens("a", "b"),
            pgens("a", "b"),
        ],
    )
    node23 = Template(
        wraptext("foo"),
        [
            pgens("a", "b"),
            pgens("c", "d"),
            pgens("e", "f"),
            pgens("a", "b"),
            pgens("a", "b"),
        ],
    )
    node24 = Template(
        wraptext("foo"),
        [
            pgens("a", "b"),
            pgens("c", "d"),
            pgens("e", "f"),
            pgens("a", "b"),
            pgens("a", "b"),
        ],
    )
    node25 = Template(
        wraptext("foo"),
        [
            pgens("a", "b"),
            pgens("c", "d"),
            pgens("e", "f"),
            pgens("a", "b"),
            pgens("a", "b"),
        ],
    )
    node26 = Template(
        wraptext("foo"),
        [
            pgens("a", "b"),
            pgens("c", "d"),
            pgens("e", "f"),
            pgens("a", "b"),
            pgens("a", "b"),
        ],
    )
    node27 = Template(wraptext("foo"), [pgenh("1", "bar")])
    node28 = Template(wraptext("foo"), [pgenh("1", "bar")])

    node2.remove("1")
    node2.remove("abc")
    node3.remove(1, keep_field=True)
    node3.remove("abc", keep_field=True)
    node4.remove("1", keep_field=False)
    node5.remove("a", keep_field=False)
    node6.remove("a", keep_field=True)
    node7.remove(1, keep_field=True)
    node8.remove(1, keep_field=False)
    node9.remove(1, keep_field=True)
    node10.remove(1, keep_field=False)
    node11.remove(node11.params[0], keep_field=False)
    node12.remove(node12.params[0], keep_field=True)
    node13.remove(node13.params[1], keep_field=False)
    node14.remove(node14.params[1], keep_field=True)
    node15.remove(node15.params[2], keep_field=False)
    node16.remove(node16.params[2], keep_field=True)
    node17.remove(node17.params[0], keep_field=False)
    node18.remove(node18.params[0], keep_field=True)
    node19.remove(node19.params[1], keep_field=False)
    node20.remove(node20.params[1], keep_field=True)
    node21.remove("a", keep_field=False)
    node22.remove("a", keep_field=True)
    node23.remove(node23.params[0], keep_field=False)
    node24.remove(node24.params[0], keep_field=True)
    node25.remove(node25.params[3], keep_field=False)
    node26.remove(node26.params[3], keep_field=True)

    with pytest.raises(ValueError):
        node1.remove(1)
    with pytest.raises(ValueError):
        node1.remove("a")
    with pytest.raises(ValueError):
        node2.remove("1")
    assert "{{foo}}" == node2
    assert "{{foo||abc=}}" == node3
    assert "{{foo|2=baz}}" == node4
    assert "{{foo|b=c}}" == node5
    assert "{{foo| a=|b=c}}" == node6
    assert "{{foo|1  =|2=c}}" == node7
    assert "{{foo|2=c}}" == node8
    assert "{{foo||c}}" == node9
    assert "{{foo|2=c}}" == node10
    assert "{{foo|b=c|a =d}}" == node11
    assert "{{foo| a=|b=c|a =d}}" == node12
    assert "{{foo| a=b|a =d}}" == node13
    assert "{{foo| a=b|b=|a =d}}" == node14
    assert "{{foo| a=b|b=c}}" == node15
    assert "{{foo| a=b|b=c|a =}}" == node16
    assert "{{foo|b|c}}" == node17
    assert "{{foo|1  =|b|c}}" == node18
    assert "{{foo|1  =a|2=c}}" == node19
    assert "{{foo|1  =a||c}}" == node20
    assert "{{foo|c=d|e=f}}" == node21
    assert "{{foo|a=|c=d|e=f}}" == node22
    assert "{{foo|c=d|e=f|a=b|a=b}}" == node23
    assert "{{foo|a=|c=d|e=f|a=b|a=b}}" == node24
    assert "{{foo|a=b|c=d|e=f|a=b}}" == node25
    assert "{{foo|a=b|c=d|e=f|a=|a=b}}" == node26
    with pytest.raises(ValueError):
        node27.remove(node28.get(1))


def test_formatting():
    """test realistic param manipulation with complex whitespace formatting
    (assumes that parsing works correctly)"""
    tests = [
        # https://en.wikipedia.org/w/index.php?title=Lamar_County,_Georgia&oldid=792356004
        (
            """{{Infobox U.S. county
| county = Lamar County
| state = Georgia
| seal =
| founded = 1920
| seat wl = Barnesville
| largest city wl = Barnesville
| area_total_sq_mi = 186
| area_land_sq_mi = 184
| area_water_sq_mi = 2.3
| area percentage = 1.3%
| census yr = 2010
| pop = 18317
| density_sq_mi = 100
| time zone = Eastern
| footnotes =
| web = www.lamarcountyga.com
| ex image = Lamar County Georgia Courthouse.jpg
| ex image cap = Lamar County courthouse in Barnesville
| district = 3rd
| named for = [[Lucius Quintus Cincinnatus Lamar II]]
}}""",
            """@@ -11,4 +11,4 @@
 | area percentage = 1.3%
-| census yr = 2010
-| pop = 18317
+| census estimate yr = 2016
+| pop = 12345<ref>example ref</ref>
 | density_sq_mi = 100""",
        ),
        # https://en.wikipedia.org/w/index.php?title=Rockdale_County,_Georgia&oldid=792359760
        (
            """{{Infobox U.S. County|
 county = Rockdale County |
 state = Georgia |
 seal =  |
 founded = October 18, 1870 |
 seat wl = Conyers |
 largest city wl = Conyers |
 area_total_sq_mi = 132 |
 area_land_sq_mi = 130 |
 area_water_sq_mi = 2.3 |
 area percentage = 1.7% |
 census yr = 2010|
 pop = 85215 |
 density_sq_mi = 657 |
 web = www.rockdalecounty.org
| ex image = Rockdale-county-courthouse.jpg
| ex image cap = Rockdale County Courthouse in Conyers
| district = 4th
| time zone= Eastern
}}""",
            """@@ -11,4 +11,4 @@
  area percentage = 1.7% |
- census yr = 2010|
- pop = 85215 |
+ census estimate yr = 2016 |
+ pop = 12345<ref>example ref</ref> |
  density_sq_mi = 657 |""",
        ),
        # https://en.wikipedia.org/w/index.php?title=Spalding_County,_Georgia&oldid=792360413
        (
            """{{Infobox U.S. County|
| county = Spalding County |
| state = Georgia |
| seal =  |
| founded = 1851 |
| seat wl = Griffin |
| largest city wl = Griffin |
| area_total_sq_mi = 200 |
| area_land_sq_mi = 196 |
| area_water_sq_mi = 3.1 |
| area percentage = 1.6% |
| census yr = 2010|
| pop = 64073 |
| density_sq_mi = 326 |
| web = www.spaldingcounty.com |
| named for = [[Thomas Spalding]]
| ex image = Spalding County Courthouse (NE corner).JPG
| ex image cap = Spalding County Courthouse in Griffin
| district = 3rd
| time zone = Eastern
}}""",
            """@@ -11,4 +11,4 @@
 | area percentage = 1.6% |
-| census yr = 2010|
-| pop = 64073 |
+|
+| census estimate yr = 2016 | pop = 12345<ref>example ref</ref> |
 | density_sq_mi = 326 |""",
        ),
        # https://en.wikipedia.org/w/index.php?title=Clinton_County,_Illinois&oldid=794694648
        (
            """{{Infobox U.S. county
 |county  = Clinton County
 |state = Illinois
| ex image           = File:Clinton County Courthouse, Carlyle.jpg
| ex image cap       = [[Clinton County Courthouse (Illinois)|Clinton County Courthouse]]
 |seal =
 |founded = 1824
 |named for = [[DeWitt Clinton]]
 |seat wl= Carlyle
| largest city wl = Breese
 |time zone=Central
 |area_total_sq_mi = 503
 |area_land_sq_mi = 474
 |area_water_sq_mi = 29
 |area percentage = 5.8%
 |census yr = 2010
 |pop = 37762
 |density_sq_mi = 80
 |web = www.clintonco.illinois.gov
| district = 15th
}}""",
            """@@ -15,4 +15,4 @@
  |area percentage = 5.8%
- |census yr = 2010
- |pop = 37762
+ |census estimate yr = 2016
+ |pop = 12345<ref>example ref</ref>
  |density_sq_mi = 80""",
        ),
        # https://en.wikipedia.org/w/index.php?title=Winnebago_County,_Illinois&oldid=789193800
        (
            """{{Infobox U.S. county |
 county  = Winnebago County |
 state = Illinois |
 seal = Winnebago County il seal.png |
 named for = [[Winnebago (tribe)|Winnebago Tribe]] |
 seat wl= Rockford |
 largest city wl = Rockford|
 area_total_sq_mi = 519 |
 area_land_sq_mi = 513|
 area_water_sq_mi = 5.9 |
 area percentage = 1.1% |
 census yr = 2010|
 pop = 295266 |
 density_sq_mi = 575
| web = www.wincoil.us
| founded year = 1836
| founded date = January 16
| time zone = Central
| district = 16th
| district2 = 17th
}}""",
            """@@ -11,4 +11,4 @@
  area percentage = 1.1% |
- census yr = 2010|
- pop = 295266 |
+ census estimate yr = 2016|
+ pop = 12345<ref>example ref</ref> |
  density_sq_mi = 575""",
        ),
    ]

    for original, expected in tests:
        code = parse(original)
        template = code.filter_templates()[0]
        template.add("pop", "12345<ref>example ref</ref>")
        template.add("census estimate yr", "2016", before="pop")
        template.remove("census yr")

        oldlines = original.splitlines(True)
        newlines = str(code).splitlines(True)
        difflines = unified_diff(oldlines, newlines, n=1)
        diff = "".join(list(difflines)[2:]).strip()
        assert expected == diff



# Copyright (C) 2012-2020 Ben Kurtovic <ben.kurtovic@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Test cases for the Argument node.
"""

import pytest

from mwparserfromhell.nodes import Argument, Text
from .conftest import assert_wikicode_equal, wrap, wraptext


def test_str():
    """test Argument.__str__()"""
    node = Argument(wraptext("foobar"))
    assert "{{{foobar}}}" == str(node)
    node2 = Argument(wraptext("foo"), wraptext("bar"))
    assert "{{{foo|bar}}}" == str(node2)


def test_children():
    """test Argument.__children__()"""
    node1 = Argument(wraptext("foobar"))
    node2 = Argument(wraptext("foo"), wrap([Text("bar"), Text("baz")]))
    gen1 = node1.__children__()
    gen2 = node2.__children__()
    assert node1.name is next(gen1)
    assert node2.name is next(gen2)
    assert node2.default is next(gen2)
    with pytest.raises(StopIteration):
        next(gen1)
    with pytest.raises(StopIteration):
        next(gen2)


def test_strip():
    """test Argument.__strip__()"""
    node1 = Argument(wraptext("foobar"))
    node2 = Argument(wraptext("foo"), wraptext("bar"))
    assert node1.__strip__() is None
    assert "bar" == node2.__strip__()


def test_showtree():
    """test Argument.__showtree__()"""
    output = []
    getter, marker = object(), object()
    get = lambda code: output.append((getter, code))
    mark = lambda: output.append(marker)
    node1 = Argument(wraptext("foobar"))
    node2 = Argument(wraptext("foo"), wraptext("bar"))
    node1.__showtree__(output.append, get, mark)
    node2.__showtree__(output.append, get, mark)
    valid = [
        "{{{",
        (getter, node1.name),
        "}}}",
        "{{{",
        (getter, node2.name),
        "    | ",
        marker,
        (getter, node2.default),
        "}}}",
    ]
    assert valid == output


def test_name():
    """test getter/setter for the name attribute"""
    name = wraptext("foobar")
    node1 = Argument(name)
    node2 = Argument(name, wraptext("baz"))
    assert name is node1.name
    assert name is node2.name
    node1.name = "héhehé"
    node2.name = "héhehé"
    assert_wikicode_equal(wraptext("héhehé"), node1.name)
    assert_wikicode_equal(wraptext("héhehé"), node2.name)


def test_default():
    """test getter/setter for the default attribute"""
    default = wraptext("baz")
    node1 = Argument(wraptext("foobar"))
    node2 = Argument(wraptext("foobar"), default)
    assert None is node1.default
    assert default is node2.default
    node1.default = "buzz"
    node2.default = None
    assert_wikicode_equal(wraptext("buzz"), node1.default)
    assert None is node2.default



# Copyright (C) 2012-2020 Ben Kurtovic <ben.kurtovic@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from ._base import Node
from ..utils import parse_anything

__all__ = ["Wikilink"]


class Wikilink(Node):
    """Represents an internal wikilink, like ``[[Foo|Bar]]``."""

    def __init__(self, title, text=None):
        super().__init__()
        self.title = title
        self.text = text

    def __str__(self):
        if self.text is not None:
            return "[[" + str(self.title) + "|" + str(self.text) + "]]"
        return "[[" + str(self.title) + "]]"

    def __children__(self):
        yield self.title
        if self.text is not None:
            yield self.text

    def __strip__(self, **kwargs):
        if self.text is not None:
            return self.text.strip_code(**kwargs)
        return self.title.strip_code(**kwargs)

    def __showtree__(self, write, get, mark):
        write("[[")
        get(self.title)
        if self.text is not None:
            write("    | ")
            mark()
            get(self.text)
        write("]]")

    @property
    def title(self):
        """The title of the linked page, as a :class:`.Wikicode` object."""
        return self._title

    @property
    def text(self):
        """The text to display (if any), as a :class:`.Wikicode` object."""
        return self._text

    @title.setter
    def title(self, value):
        self._title = parse_anything(value)

    @text.setter
    def text(self, value):
        if value is None:
            self._text = None
        else:
            self._text = parse_anything(value)



# Copyright (C) 2012-2020 Ben Kurtovic <ben.kurtovic@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
This module contains the :class:`.StringMixIn` type, which implements the
interface for the ``str`` type in a dynamic manner.
"""

from sys import getdefaultencoding

__all__ = ["StringMixIn"]


def inheritdoc(method):
    """Set __doc__ of *method* to __doc__ of *method* in its parent class.

    Since this is used on :class:`.StringMixIn`, the "parent class" used is
    ``str``. This function can be used as a decorator.
    """
    method.__doc__ = getattr(str, method.__name__).__doc__
    return method


class StringMixIn:
    """Implement the interface for ``str`` in a dynamic manner.

    To use this class, inherit from it and override the :meth:`__str__` method
    to return the string representation of the object. The various string
    methods will operate on the value of :meth:`__str__` instead of the
    immutable ``self`` like the regular ``str`` type.
    """

    def __str__(self):
        raise NotImplementedError()

    def __bytes__(self):
        return bytes(self.__str__(), getdefaultencoding())

    def __repr__(self):
        return repr(self.__str__())

    def __lt__(self, other):
        return self.__str__() < other

    def __le__(self, other):
        return self.__str__() <= other

    def __eq__(self, other):
        return self.__str__() == other

    def __ne__(self, other):
        return self.__str__() != other

    def __gt__(self, other):
        return self.__str__() > other

    def __ge__(self, other):
        return self.__str__() >= other

    def __bool__(self):
        return bool(self.__str__())

    def __len__(self):
        return len(self.__str__())

    def __iter__(self):
        yield from self.__str__()

    def __getitem__(self, key):
        return self.__str__()[key]

    def __reversed__(self):
        return reversed(self.__str__())

    def __contains__(self, item):
        return str(item) in self.__str__()

    def __getattr__(self, attr):
        if not hasattr(str, attr):
            raise AttributeError(
                "{!r} object has no attribute {!r}".format(type(self).__name__, attr)
            )
        return getattr(self.__str__(), attr)

    maketrans = str.maketrans  # Static method can't rely on __getattr__


del inheritdoc



#! /usr/bin/env python
#
# Copyright (C) 2012-2023 Ben Kurtovic <ben.kurtovic@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from glob import glob
import os
import sys

from setuptools import find_packages, setup, Extension
from setuptools.command.build_ext import build_ext

sys.path.insert(0, os.path.join(os.path.dirname(__file__), "src"))

from mwparserfromhell import __version__

with open("README.rst") as fp:
    long_docs = fp.read()

use_extension = True
fallback = True

# Allow env var WITHOUT_EXTENSION and args --with[out]-extension:

env_var = os.environ.get("WITHOUT_EXTENSION")
if "--without-extension" in sys.argv:
    use_extension = False
elif "--with-extension" in sys.argv:
    fallback = False
elif env_var is not None:
    if env_var == "1":
        use_extension = False
    elif env_var == "0":
        fallback = False

# Remove the command line argument as it isn't understood by setuptools:

sys.argv = [
    arg for arg in sys.argv if arg not in ("--without-extension", "--with-extension")
]


def build_ext_patched(self):
    try:
        build_ext_original(self)
    except Exception as exc:
        print("error: " + str(exc))
        print("Falling back to pure Python mode.")
        del self.extensions[:]


if fallback:
    build_ext.run, build_ext_original = build_ext_patched, build_ext.run

# Project-specific part begins here:

tokenizer = Extension(
    "mwparserfromhell.parser._tokenizer",
    sources=sorted(glob("src/mwparserfromhell/parser/ctokenizer/*.c")),
    depends=sorted(glob("src/mwparserfromhell/parser/ctokenizer/*.h")),
)

setup(
    name="mwparserfromhell",
    packages=find_packages("src"),
    package_dir={"": "src"},
    ext_modules=[tokenizer] if use_extension else [],
    setup_requires=["pytest-runner"]
    if "test" in sys.argv or "pytest" in sys.argv
    else [],
    tests_require=["pytest"],
    version=__version__,
    python_requires=">= 3.7",
    author="Ben Kurtovic",
    author_email="ben.kurtovic@gmail.com",
    url="https://github.com/earwig/mwparserfromhell",
    description="MWParserFromHell is a parser for MediaWiki wikicode.",
    long_description=long_docs,
    download_url="https://github.com/earwig/mwparserfromhell/tarball/v{}".format(
        __version__
    ),
    keywords="earwig mwparserfromhell wikipedia wiki mediawiki wikicode template parsing",
    license="MIT License",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Text Processing :: Markup",
    ],
)



# Copyright (C) 2012-2020 Ben Kurtovic <ben.kurtovic@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Integration test cases for mwparserfromhell's documentation.
"""

import json
from io import StringIO
import os
from urllib.parse import urlencode
from urllib.request import urlopen

import pytest

import mwparserfromhell


def assert_print(value, output):
    """Assertion check that *value*, when printed, produces *output*."""
    buff = StringIO()
    print(value, end="", file=buff)
    buff.seek(0)
    assert output == buff.read()


def test_readme_1():
    """test a block of example code in the README"""
    text = "I has a template! {{foo|bar|baz|eggs=spam}} See it?"
    wikicode = mwparserfromhell.parse(text)
    assert_print(wikicode, "I has a template! {{foo|bar|baz|eggs=spam}} See it?")
    templates = wikicode.filter_templates()
    assert_print(templates, "['{{foo|bar|baz|eggs=spam}}']")
    template = templates[0]
    assert_print(template.name, "foo")
    assert_print(template.params, "['bar', 'baz', 'eggs=spam']")
    assert_print(template.get(1).value, "bar")
    assert_print(template.get("eggs").value, "spam")


def test_readme_2():
    """test a block of example code in the README"""
    text = "{{foo|{{bar}}={{baz|{{spam}}}}}}"
    temps = mwparserfromhell.parse(text).filter_templates()
    res = "['{{foo|{{bar}}={{baz|{{spam}}}}}}', '{{bar}}', '{{baz|{{spam}}}}', '{{spam}}']"
    assert_print(temps, res)


def test_readme_3():
    """test a block of example code in the README"""
    code = mwparserfromhell.parse("{{foo|this {{includes a|template}}}}")
    assert_print(
        code.filter_templates(recursive=False),
        "['{{foo|this {{includes a|template}}}}']",
    )
    foo = code.filter_templates(recursive=False)[0]
    assert_print(foo.get(1).value, "this {{includes a|template}}")
    assert_print(foo.get(1).value.filter_templates()[0], "{{includes a|template}}")
    assert_print(foo.get(1).value.filter_templates()[0].get(1).value, "template")


def test_readme_4():
    """test a block of example code in the README"""
    text = "{{cleanup}} '''Foo''' is a [[bar]]. {{uncategorized}}"
    code = mwparserfromhell.parse(text)
    for template in code.filter_templates():
        if template.name.matches("Cleanup") and not template.has("date"):
            template.add("date", "July 2012")
    res = "{{cleanup|date=July 2012}} '''Foo''' is a [[bar]]. {{uncategorized}}"
    assert_print(code, res)
    code.replace("{{uncategorized}}", "{{bar-stub}}")
    res = "{{cleanup|date=July 2012}} '''Foo''' is a [[bar]]. {{bar-stub}}"
    assert_print(code, res)
    res = "['{{cleanup|date=July 2012}}', '{{bar-stub}}']"
    assert_print(code.filter_templates(), res)
    text = str(code)
    res = "{{cleanup|date=July 2012}} '''Foo''' is a [[bar]]. {{bar-stub}}"
    assert_print(text, res)
    assert text == code


@pytest.mark.skipif("NOWEB" in os.environ, reason="web test disabled by environ var")
def test_readme_5():
    """test a block of example code in the README; includes a web call"""
    url1 = "https://en.wikipedia.org/w/api.php"
    url2 = "https://en.wikipedia.org/w/index.php?title={0}&action=raw"
    title = "Test"
    data = {
        "action": "query",
        "prop": "revisions",
        "rvprop": "content",
        "rvslots": "main",
        "rvlimit": 1,
        "titles": title,
        "format": "json",
        "formatversion": "2",
    }
    try:
        raw = urlopen(url1, urlencode(data).encode("utf8")).read()
    except OSError:
        pytest.skip("cannot continue because of unsuccessful web call")
    res = json.loads(raw.decode("utf8"))
    revision = res["query"]["pages"][0]["revisions"][0]
    text = revision["slots"]["main"]["content"]
    try:
        expected = urlopen(url2.format(title)).read().decode("utf8")
    except OSError:
        pytest.skip("cannot continue because of unsuccessful web call")
    actual = mwparserfromhell.parse(text)
    assert expected == actual
